/*****************************************************************************/

/*
 *      dabusb.c  --  DAB USB & DR-BOX 1 driver
 *      
 *      Usable for the DAB-USB receiver of BayCom (4 keys) and
 *      for Terratec's DR BOX 1 (12 keys)
 *
 *      Copyright (C) 1999-2004 Deti Fliegl, Georg Acher, BayCom GmbH, www.baycom.de
 *
 *      The firmware (CPU/FPGA) is contained in dabfirmware.h 
 *      and works with both receivers, but only the keyboard
 *      of the DR BOX 1 (12 keys) is supported with this firmware.
 *      If you want to use the 4 keys of BayCom's DAB-USB, get the 
 *      appropriate firmware at:  http://www.baycom.de
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *
 *  $Id: dabusb.c,v 1.1.2.2 2010/09/13 14:42:34 fliegl Exp $
 *
 */

/*****************************************************************************/

#include <linux/module.h>
#include <linux/socket.h>
#include <linux/list.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <asm/atomic.h>
#include <linux/delay.h>
#include <linux/usb.h>
#include <linux/smp_lock.h>
#include <linux/poll.h>
#include <linux/pci.h>
#include <linux/device.h>
//#define DEBUG 1
#include "dabusb.h"
#include "dabfirmware.h"

/*
#define USE_4KEY_FIRMWARE
*/

#ifdef USE_4KEY_FIRMWARE
#include "dabfirmware_4.h"
#endif

/*
 * Version Information
 */
#define DRIVER_VERSION "v2.00"
#define DRIVER_AUTHOR "Deti Fliegl, Georg Acher, BayCom GmbH, www.baycom.de"
#define DRIVER_DESC "DAB-USB/DR-BOX1 Interface Driver for Linux (c)1999-2004"


/* --------------------------------------------------------------------- */

#ifdef CONFIG_USB_DYNAMIC_MINORS
#define NRDABUSB 256
#else
#define NRDABUSB 4
#endif

/*-------------------------------------------------------------------*/

static DECLARE_MUTEX (disconnect_sem);
static int buffers = 256;


static int dabusb_probe (struct usb_interface *intf,
                           const struct usb_device_id *id);
static void dabusb_disconnect (struct usb_interface *intf);
static ssize_t dabusb_read (struct file *file, char *buf, size_t count, loff_t * ppos);
static int dabusb_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
static int dabusb_open (struct inode *inode, struct file *file);
static int dabusb_release (struct inode *inode, struct file *file);
static unsigned int dabusb_poll(struct file *file, struct poll_table_struct *wait);

static struct usb_device_id dabusb_ids [] = {
	{ USB_DEVICE(0x0ccd, 0x0001) },       /* Terratec/BayCom DAB receiver PC (12 keys) (dumb)*/
	{ USB_DEVICE(0x0ccd, 0x8001) },       /* Terratec/BayCom DAB receiver PC (12 keys) (downloaded) */
	{ USB_DEVICE(0x0ccd, 0x0002) },       /* Terratec DAB Receiver HiFi (dumb) */
	{ USB_DEVICE(0x0ccd, 0x8002) },       /* Terratec DAB Receiver HiFi (downloaded) */
#if 1
	{ USB_DEVICE(0x0547, 0x2131) },       /* BayCom DABUSB prototype (4 keys)(dumb) */
	{ USB_DEVICE(0x0547, 0x9999) },       /* BayCom DABUSB prototype (4 keys) (downloaded */
#endif
	{ }						/* Terminating entry */
};

MODULE_DEVICE_TABLE (usb, dabusb_ids);

static struct usb_driver dabusb_driver = {
	.name =		"dabusb",
	.probe =	dabusb_probe,
	.disconnect =	dabusb_disconnect,
	.id_table =	dabusb_ids,
};

/*-------------------------------------------------------------------*/

static int dabusb_add_buf_tail (pdabusb_t s, struct list_head *dst, struct list_head *src)
{
	unsigned long flags;
	struct list_head *tmp;
	int ret = 0;

	spin_lock_irqsave (&s->lock, flags);

	if (list_empty (src)) {
		// no elements in source buffer
		ret = -1;
		goto err;
	}
	tmp = src->next;
	list_del (tmp);
	list_add_tail (tmp, dst);

  err:	spin_unlock_irqrestore (&s->lock, flags);
	return ret;
}
/*-------------------------------------------------------------------*/
#ifdef DEBUG 
static void dump_urb (struct urb* purb)
{
	dev_err(&s->usbdev->dev, "urb                   :%p", purb);
	dev_err(&s->usbdev->dev, "dev                   :%p", purb->dev);
	dev_err(&s->usbdev->dev, "pipe                  :%08X", purb->pipe);
	dev_err(&s->usbdev->dev, "status                :%d", purb->status);
	dev_err(&s->usbdev->dev, "transfer_flags        :%08X", purb->transfer_flags);
	dev_err(&s->usbdev->dev, "transfer_buffer       :%p", purb->transfer_buffer);
	dev_err(&s->usbdev->dev, "transfer_buffer_length:%d", purb->transfer_buffer_length);
	dev_err(&s->usbdev->dev, "actual_length         :%d", purb->actual_length);
	dev_err(&s->usbdev->dev, "setup_packet          :%p", purb->setup_packet);
	dev_err(&s->usbdev->dev, "start_frame           :%d", purb->start_frame);
	dev_err(&s->usbdev->dev, "number_of_packets     :%d", purb->number_of_packets);
	dev_err(&s->usbdev->dev, "interval              :%d", purb->interval);
	dev_err(&s->usbdev->dev, "error_count           :%d", purb->error_count);
	dev_err(&s->usbdev->dev, "context               :%p", purb->context);
	dev_err(&s->usbdev->dev, "complete              :%p", purb->complete);
}
#endif
/*-------------------------------------------------------------------*/
static int dabusb_cancel_queue (pdabusb_t s, struct list_head *q)
{
	unsigned long flags;
	struct list_head *p;
	pbuff_t b;

	dbg("dabusb_cancel_queue");

	spin_lock_irqsave (&s->lock, flags);

	for (p = q->next; p != q; p = p->next) {
		b = list_entry (p, buff_t, buff_list);

#ifdef DEBUG
		dump_urb(b->purb);
#endif
		usb_kill_urb (b->purb);
	}
	spin_unlock_irqrestore (&s->lock, flags);
	return 0;
}
/*-------------------------------------------------------------------*/
static int dabusb_free_queue (struct list_head *q)
{
	struct list_head *tmp;
	struct list_head *p;
	pbuff_t b;

	dbg("dabusb_free_queue");
	for (p = q->next; p != q;) {
		b = list_entry (p, buff_t, buff_list);

#ifdef DEBUG 
		dump_urb(b->purb);
#endif
		usb_free_urb(b->purb);
		tmp = p->next;
		list_del (p);
		kfree (b);
		p = tmp;
	}

	return 0;
}
/*-------------------------------------------------------------------*/
static int dabusb_free_buffers (pdabusb_t s)
{
	unsigned long flags;
	dbg("dabusb_free_buffers");
	s->got_mem = 0;

	spin_lock_irqsave(&s->lock, flags);

	dabusb_free_queue (&s->free_buff_list);
	dabusb_free_queue (&s->rec_buff_list);

	spin_unlock_irqrestore(&s->lock, flags);
        if (s->rx_bufs){
	        dev_warn(&s->usbdev->dev, "dabusb_free_buffers: pci_free_consistent\n");
        	pci_free_consistent(NULL, s->total_buffer_size , s->rx_bufs, s->rx_bufs_dma);
        	s->rx_bufs=NULL;
	}
                                        
	return 0;
}
/*-------------------------------------------------------------------*/
static void dabusb_iso_complete (struct urb *purb)
{
	pbuff_t b = purb->context;
	pdabusb_t s = b->s;
	int i;
	int len;
	int dst = 0;
	void *buf = purb->transfer_buffer;

	dbg("dabusb_iso_complete");

	// process if URB was not killed
	if (purb->status != -ENOENT) {
		unsigned int pipe = usb_rcvisocpipe (s->usbdev, _DABUSB_ISOPIPE);
		int pipesize = usb_maxpacket (s->usbdev, pipe, usb_pipeout (pipe));
		for (i = 0; i < purb->number_of_packets; i++)
			if (!purb->iso_frame_desc[i].status) {
				len = purb->iso_frame_desc[i].actual_length;
				if (len <= pipesize) {
					if (len&3)
						dev_err(&s->usbdev->dev, "dabusb_iso_complete: length not a multiple of 4 (%i)",len);
					len=len&0xffc; /* only multiples of 4, bugfix for SiS/OHCI */

					memcpy (buf + dst, buf + purb->iso_frame_desc[i].offset, len);
					dst += len;
				}
				else
					dev_err(&s->usbdev->dev, "dabusb_iso_complete: invalid len %d", len);
			}
			else
				dev_warn(&s->usbdev->dev, "dabusb_iso_complete: corrupted packet status: %d", purb->iso_frame_desc[i].status);

		if (purb->actual_length&3)
		{
			dev_err(&s->usbdev->dev, "dabusb_iso_complete: purb->actual_length not a multiple of 4 (%i)",purb->actual_length);
			purb->actual_length&=~3; /* only multiples of 4, bugfix for SiS/OHCI */
		}
		if (dst != purb->actual_length)
			dev_err(&s->usbdev->dev, "dst!=purb->actual_length:%d!=%d", dst, purb->actual_length);
	}

	if (atomic_dec_and_test (&s->pending_io) && !s->remove_pending && s->state != _stopped) {
		s->overruns++;
		dev_err(&s->usbdev->dev, "overrun (%d)", s->overruns);
	}
	wake_up (&s->wait);
}
/*-------------------------------------------------------------------*/
static int dabusb_alloc_buffers (pdabusb_t s)
{
	int buffers = 0;
	pbuff_t b;
	unsigned int pipe = usb_rcvisocpipe (s->usbdev, _DABUSB_ISOPIPE);
	int pipesize = usb_maxpacket (s->usbdev, pipe, usb_pipeout (pipe));
	int packets;
	int transfer_buffer_length;
	int i;
	
	if(!pipesize)
	{
		dev_err(&s->usbdev->dev, "ISO-pipe has size 0!!");
		return -ENOMEM;
	}
	packets = _ISOPIPESIZE / pipesize;
	transfer_buffer_length = packets * pipesize;
	
	dev_warn(&s->usbdev->dev, "dabusb_alloc_buffers pipesize:%d packets:%d transfer_buffer_len:%d",
		 pipesize, packets, transfer_buffer_length);

        s->rx_bufs = pci_alloc_consistent(NULL, s->total_buffer_size, &s->rx_bufs_dma);
        
        dev_warn(&s->usbdev->dev, "dabusb_alloc_buffers pci_alloc_consistent done\n");
        if (s->rx_bufs == NULL) {
        	dev_err(&s->usbdev->dev, "pci_alloc_consistent==NULL");
        	return 0;
	}
       
	while (buffers < s->total_buffer_size) {
		b = (pbuff_t) kmalloc (sizeof (buff_t), GFP_KERNEL);
		if (!b) {
			dev_err(&s->usbdev->dev, "kmalloc(sizeof(buff_t))==NULL");
			goto err;
		}
		memset (b, 0, sizeof (buff_t));
		b->s = s;
		b->purb = usb_alloc_urb(packets,GFP_ATOMIC);
		if (!b->purb) {
			dev_err(&s->usbdev->dev, "usb_alloc_urb == NULL");
			kfree (b);
			goto err;
		}

		b->purb->transfer_buffer = s->rx_bufs+buffers;
		b->purb->transfer_buffer_length = transfer_buffer_length;
		b->purb->number_of_packets = packets;
		b->purb->complete = dabusb_iso_complete;
		b->purb->context = b;
		b->purb->dev = s->usbdev;
		b->purb->pipe = pipe;
		b->purb->interval = 1;
		b->purb->transfer_flags = URB_ISO_ASAP;

		for (i = 0; i < packets; i++) {
			b->purb->iso_frame_desc[i].offset = i * pipesize;
			b->purb->iso_frame_desc[i].length = pipesize;
		}

		buffers += transfer_buffer_length;
		list_add_tail (&b->buff_list, &s->free_buff_list);
	}
	s->got_mem=1;

	return 0;

	err:
	dabusb_free_buffers (s);
	return -ENOMEM;
}
/*-------------------------------------------------------------------*/
static int dabusb_bulk (pdabusb_t s, pbulk_transfer_t pb)
{
	int ret;
	unsigned int pipe;
	int actual_length;

	dbg("dabusb_bulk");

	if (!pb->pipe)
		pipe = usb_rcvbulkpipe (s->usbdev, 2);
	else
		pipe = usb_sndbulkpipe (s->usbdev, 2);

	ret=usb_bulk_msg(s->usbdev, pipe, pb->data, pb->size, &actual_length, 100);
#if 1
	if(ret<0) {
		dev_err(&s->usbdev->dev, "dabusb: usb_bulk_msg failed(%d)",ret);
#if 0
		if (usb_set_interface (s->usbdev, _DABUSB_IF, 1) < 0) {
			dev_err(&s->usbdev->dev, "set_interface failed");
			return -EINVAL;
		}
#endif		
	}	
#endif

	if( ret == -EPIPE ) {
		dev_warn(&s->usbdev->dev, "CLEAR_FEATURE request to remove STALL condition.");
		if(usb_clear_halt(s->usbdev, usb_pipeendpoint(pipe)))
			dev_err(&s->usbdev->dev, "request failed");
	}

	pb->size = actual_length;
	return ret;
}
/* --------------------------------------------------------------------- */
static int dabusb_writemem (pdabusb_t s, int pos, unsigned char *data, int len)
{
	int ret;
	unsigned char *transfer_buffer =  kmalloc (len, GFP_KERNEL);

	if (!transfer_buffer) {
		dev_err(&s->usbdev->dev, "dabusb_writemem: kmalloc(%d) failed.", len);
		return -ENOMEM;
	}

	memcpy (transfer_buffer, data, len);

	ret=usb_control_msg(s->usbdev, usb_sndctrlpipe( s->usbdev, 0 ), 0xa0, 0x40, pos, 0, transfer_buffer, len, 300);

	kfree (transfer_buffer);
	return ret;
}
/* --------------------------------------------------------------------- */
static int dabusb_8051_reset (pdabusb_t s, unsigned char reset_bit)
{
	dbg("dabusb_8051_reset: %d",reset_bit);
	return dabusb_writemem (s, CPUCS_REG, &reset_bit, 1);
}
/* --------------------------------------------------------------------- */
static int dabusb_loadmem (pdabusb_t s, int fw_type)
{
	int ret;
	PINTEL_HEX_RECORD ptr = firmware;

#ifdef USE_4KEY_FIRMWARE
	if (fw_type==1)
		ptr = firmware4;
#endif
	dbg("Enter dabusb_loadmem (internal)");
	
	ret = dabusb_8051_reset (s, 1);
	while (ptr->Type == 0) {
		dbg("dabusb_writemem: %04X %p %d)", ptr->Address, ptr->Data, ptr->Length);

		ret = dabusb_writemem (s, ptr->Address, ptr->Data, ptr->Length);
		if (ret < 0) {
			dev_err(&s->usbdev->dev, "dabusb_writemem failed (%d %04X %p %d)", ret, ptr->Address, ptr->Data, ptr->Length);
			break;
		}
		mdelay(1);
		ptr++;
	}
	dbg("dabusb_loadmem: reset");
	ret = dabusb_8051_reset (s, 0);
	
	dbg("dabusb_loadmem: exit");

	return ret;
}
/* --------------------------------------------------------------------- */
static int dabusb_fpga_clear (pdabusb_t s, pbulk_transfer_t b)
{
	b->size = 4;
	b->data[0] = 0x2a;
	b->data[1] = 0;
	b->data[2] = 0;
	b->data[3] = 0;

	dbg("dabusb_fpga_clear");

	return dabusb_bulk (s, b);
}
/* --------------------------------------------------------------------- */
static int dabusb_fpga_init (pdabusb_t s, pbulk_transfer_t b)
{
	b->size = 4;
	b->data[0] = 0x2c;
	b->data[1] = 0;
	b->data[2] = 0;
	b->data[3] = 0;

	dev_err(&s->usbdev->dev, "dabusb_fpga_init");

	return dabusb_bulk (s, b);
}
/* --------------------------------------------------------------------- */
static int dabusb_fpga_download (pdabusb_t s, const char *fname)
{
	pbulk_transfer_t b = kmalloc (sizeof (bulk_transfer_t), GFP_KERNEL);
	unsigned int blen, n;
	int ret,retry;
	int download_ok;
	unsigned char *buf = bitstream;

	dbg("Enter dabusb_fpga_download (internal)");

	if (!b) {
		dev_err(&s->usbdev->dev, "kmalloc(sizeof(bulk_transfer_t))==NULL");
		return -ENOMEM;
	}

	for(retry=0;retry<3;retry++)
	{
		/* Sync endpoints with dummy command */
		b->pipe = 1;
		b->size=4;
		b->data[0]=0x11;
		b->data[1]=0;
		b->data[2]=0x84;
		b->data[3]=2;
		b->data[4]=0;
		b->data[5]=0;
		ret = dabusb_bulk (s, b);
		dbg("dabusb_fpga: sync %i\n",ret);

		b->pipe = 1;		
		ret = dabusb_fpga_clear (s, b);
		mdelay (20);
		blen = buf[73] + (buf[72] << 8);
		
		dbg("Bitstream len: %i", blen);
		
		b->data[0] = 0x2b;
		b->data[1] = 0;
		b->data[2] = 0;
		b->data[3] = 60;
		download_ok=1;

		for (n = 0; n <= blen + 60; n += 60) {
			// some cclks for startup
			b->size = 64;
			memcpy (b->data + 4, buf + 74 + n, 60);
			ret = dabusb_bulk (s, b);
			if (ret < 0) {
				dev_err(&s->usbdev->dev, "dabusb_bulk (#%i) failed (err=%i).",n,ret);
				download_ok=0;
				break;
			}
			mdelay(2);
		}
		if (download_ok)
			break;
	}

	for(retry=0;retry<3;retry++)
		if (!dabusb_fpga_init (s, b))
			break;

	kfree (b);

	dbg("exit dabusb_fpga_download");

	return ret;
}

static int dabusb_stop (pdabusb_t s)
{
	dbg("dabusb_stop");

	dabusb_cancel_queue (s, &s->rec_buff_list);
	s->state = _stopped;

	dbg("pending_io: %d", s->pending_io.counter);

	s->pending_io.counter = 0;
	return 0;
}

static int dabusb_startrek (pdabusb_t s)
{
	if (!s->got_mem && s->state != _started) {

		dbg("dabusb_startrek");

		dabusb_stop (s);
		s->state = _started;
		s->readptr = 0;
	}

	if (!list_empty (&s->free_buff_list)) {
		pbuff_t end;
		int ret;
		
		while (!dabusb_add_buf_tail (s, &s->rec_buff_list, &s->free_buff_list)) {

			dbg("submitting: end:%p s->rec_buff_list:%p", s->rec_buff_list.prev, &s->rec_buff_list);

			end = list_entry (s->rec_buff_list.prev, buff_t, buff_list);

			end->purb->dev=s->usbdev;
			end->purb->status=0;
			ret = usb_submit_urb (end->purb, GFP_ATOMIC);
			if (ret) {
//				dump_urb(end->purb);
				dev_err(&s->usbdev->dev, "usb_submit_urb returned:%d", ret);
				if (dabusb_add_buf_tail (s, &s->free_buff_list, &s->rec_buff_list))
					dev_err(&s->usbdev->dev, "startrek: dabusb_add_buf_tail failed");
				break;
			}
			else
				atomic_inc (&s->pending_io);
		}
		dbg("pending_io: %d",s->pending_io.counter);
	}

	return 0;
}

static ssize_t dabusb_read (struct file *file, char *buf, size_t count, loff_t * ppos)
{
	pdabusb_t s = (pdabusb_t) file->private_data;
	unsigned long flags;
	unsigned ret = 0;
	int rem;
	int cnt;
	pbuff_t b;
	struct urb *purb = NULL;

	dbg("dabusb_read");

	if (*ppos)
		return -ESPIPE;

	if (s->remove_pending)
		return -EIO;


	if (!s->usbdev)
		return -EIO;

//	down(&s->mutex);

	while (count > 0) {
		dabusb_startrek (s);

		spin_lock_irqsave (&s->lock, flags);

		if (list_empty (&s->rec_buff_list)) {

			spin_unlock_irqrestore(&s->lock, flags);

			dev_err(&s->usbdev->dev, "error: rec_buf_list is empty");
			goto err;
		}
		
		b = list_entry (s->rec_buff_list.next, buff_t, buff_list);
		purb = b->purb;

		spin_unlock_irqrestore(&s->lock, flags);

		if (purb->status == -EINPROGRESS) {
			if (file->f_flags & O_NONBLOCK)		// return nonblocking
			 {
				if (!ret)
					ret = -EAGAIN;
				goto err;
			}

			interruptible_sleep_on (&s->wait);

			if (signal_pending (current)) {
				if (!ret)
					ret = -ERESTARTSYS;
				goto err;
			}

			spin_lock_irqsave (&s->lock, flags);

			if (list_empty (&s->rec_buff_list)) {
				spin_unlock_irqrestore(&s->lock, flags);
				dev_err(&s->usbdev->dev, "error: still no buffer available.");
				goto err;
			}
			spin_unlock_irqrestore(&s->lock, flags);
			s->readptr = 0;
		}
		if (s->remove_pending) {
			ret = -EIO;
			goto err;
		}

		rem = purb->actual_length - s->readptr;		// set remaining bytes to copy

		if (count >= rem)
			cnt = rem;
		else
			cnt = count;

		dbg("copy_to_user:%p %p %d",buf, purb->transfer_buffer + s->readptr, cnt);

		if (copy_to_user (buf, purb->transfer_buffer + s->readptr, cnt)) {
			dev_err(&s->usbdev->dev, "read: copy_to_user failed");
			if (!ret)
				ret = -EFAULT;
			goto err;
		}

		s->readptr += cnt;
		count -= cnt;
		buf += cnt;
		ret += cnt;

		if (s->readptr == purb->actual_length) {
			// finished, take next buffer
			if (dabusb_add_buf_tail (s, &s->free_buff_list, &s->rec_buff_list))
				dev_err(&s->usbdev->dev, "read: dabusb_add_buf_tail failed");
			s->readptr = 0;
		}
	}

err:	//up(&s->mutex);
	return ret;
}

static int dabusb_open (struct inode *inode, struct file *file)
{
	pdabusb_t s;
	struct usb_interface *interface;
	int retval=0;
	int subminor;

	subminor = iminor(inode);
	down(&disconnect_sem);
	
	interface = usb_find_interface (&dabusb_driver, subminor);
        if (!interface) {
		retval=-ENODEV;
		goto exit_no_device;
        }
        
        s=usb_get_intfdata(interface);
	if (!s) {
		retval=-ENODEV;
		goto exit_no_device;
	}

	dbg("dabusb_open %p",s);
	
	down (&s->mutex);
	while (!s->usbdev || s->opened) {
		up (&s->mutex);

		if (file->f_flags & O_NONBLOCK) {
			retval=-EBUSY;
			goto exit_no_device;
		}
		schedule_timeout (HZ / 2);

		if (signal_pending (current)) {
			retval= -EAGAIN;
			goto exit_no_device;
		}
		down (&s->mutex);
	}

	if (usb_set_interface (s->usbdev, _DABUSB_IF, 1) < 0) {
		dev_err(&s->usbdev->dev, "set_interface failed");
		up (&s->mutex);
		retval=-EINVAL;
		goto exit_no_device;
	}
	s->opened = 1;
	up (&s->mutex);
	file->f_pos = 0;
	file->private_data = s;

exit_no_device:
	up (&disconnect_sem);
        return retval;
}

static int dabusb_release (struct inode *inode, struct file *file)
{
	pdabusb_t s = (pdabusb_t) file->private_data;

	dbg("dabusb_release");

	down (&s->mutex);
	dabusb_stop (s);
	up (&s->mutex);

	if (!s->remove_pending) {
		if (usb_set_interface (s->usbdev, _DABUSB_IF, 0) < 0)
			dev_err(&s->usbdev->dev, "set_interface failed");
	}
	else
		wake_up (&s->remove_ok);

	s->opened = 0;
	return 0;
}

static int dabusb_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	pdabusb_t s = (pdabusb_t) file->private_data;
	pbulk_transfer_t pbulk;
	int ret = 0;
	int version = DABUSB_VERSION;

	dbg("dabusb_ioctl");

	if (s->remove_pending)
		return -EIO;

	down (&s->mutex);

	if (!s->usbdev) {
		up (&s->mutex);
		return -EIO;
	}

	switch (cmd) {

	case IOCTL_DAB_BULK:
		pbulk = (pbulk_transfer_t) kmalloc (sizeof (bulk_transfer_t), GFP_KERNEL);

		if (!pbulk) {
			ret = -ENOMEM;
			break;
		}

		if (copy_from_user (pbulk, (void *) arg, sizeof (bulk_transfer_t))) {
			ret = -EFAULT;
			kfree (pbulk);
			break;
		}

		ret=dabusb_bulk (s, pbulk);
		if(ret==0)
			ret = copy_to_user ((void *) arg, pbulk, sizeof (bulk_transfer_t));
		kfree (pbulk);
		break;
	
	case IOCTL_DAB_REMOVED:
		ret = put_user (s->remove_pending, (unsigned int *) arg);
		break;

	case IOCTL_DAB_OVERRUNS:
		ret = put_user (s->overruns, (unsigned int *) arg);
		break;

	case IOCTL_DAB_VERSION:
		ret = put_user (version, (unsigned int *) arg);
		break;

	default:
		ret = -ENOIOCTLCMD;
		break;
	}
	up (&s->mutex);
	return ret;
}

static unsigned int dabusb_poll(struct file *file, struct poll_table_struct *wait)
{
	pdabusb_t s = (pdabusb_t) file->private_data;
	unsigned long flags;
	unsigned int mask = 0;

	if (file->f_mode & FMODE_READ) {
		poll_wait(file, &s->wait, wait);
	}
	spin_lock_irqsave(&s->lock, flags);
	if (file->f_mode & FMODE_READ) {
		if (!list_empty (&s->rec_buff_list))
			mask |= POLLIN | POLLRDNORM;
	}
	spin_unlock_irqrestore(&s->lock, flags);

	return mask;
}

static const struct file_operations dabusb_fops =
{
	.owner =	THIS_MODULE,
	.llseek =	no_llseek,
	.read =		dabusb_read,
	.poll =		dabusb_poll,
	.ioctl =	dabusb_ioctl,
	.open =		dabusb_open,
	.release =	dabusb_release,
};

static struct usb_class_driver dabusb_class = {
	.name =		"dabusb%d",
	.fops =		&dabusb_fops,
	.minor_base =	DABUSB_MINOR,
};


/* --------------------------------------------------------------------- */
static int dabusb_probe (struct usb_interface *intf,
			   const struct usb_device_id *id)
{
	struct usb_device *usbdev = interface_to_usbdev(intf);
	int retval=-ENODEV;
	pdabusb_t s;

	printk(KERN_INFO "dabusb: probe: vendor id 0x%x, device id 0x%x ifnum:%d",
	     usbdev->descriptor.idVendor, usbdev->descriptor.idProduct, intf->altsetting->desc.bInterfaceNumber);

	if (intf->altsetting->desc.bInterfaceNumber != _DABUSB_IF)
		switch(usbdev->descriptor.idProduct) {
			case 0x8001:
			case 0x8002:
			case 0x9999:
				return -ENODEV;
	}
	
	s= (pdabusb_t)kmalloc(sizeof(dabusb_t),GFP_KERNEL);
	if (!s) 
		 return -ENOMEM;
	memset(s,0,sizeof(dabusb_t));
	init_MUTEX (&s->mutex);
	s->total_buffer_size = buffers<<10;
	init_waitqueue_head (&s->wait);
	init_waitqueue_head (&s->remove_ok);
	spin_lock_init (&s->lock);
	INIT_LIST_HEAD (&s->free_buff_list);
	INIT_LIST_HEAD (&s->rec_buff_list);
		
	s->remove_pending = 0;
	s->usbdev = usbdev;
	s->devnum = intf->minor;

	if (usb_reset_configuration (usbdev) < 0) {
		dev_err(&s->usbdev->dev, "reset_configuration failed");
		goto probe_failed;
	}

	if (usbdev->descriptor.idProduct  == 0x0001 ||
	    usbdev->descriptor.idProduct  == 0x0002 || 
	    usbdev->descriptor.idProduct  == 0x2131 ) {
		dabusb_loadmem (s, ( usbdev->descriptor.idProduct==0x2131?1:0));
		goto probe_failed;
	}
	else if (usbdev->descriptor.idProduct  == 0x8001 ||
		usbdev->descriptor.idProduct  == 0x8002 ||
		usbdev->descriptor.idProduct  == 0x9999) {
		dabusb_fpga_download (s, NULL);

		if (usb_set_interface (s->usbdev, _DABUSB_IF, 0) < 0) {
			dev_err(&s->usbdev->dev, "set_interface failed");
			goto probe_failed;
		}
	}

	usb_set_intfdata (intf, s);


	retval = usb_register_dev(intf, &dabusb_class);
	s->devnum = intf->minor;
	
        if (retval) {
                usb_set_intfdata (intf, NULL);
                kfree(s);
                return -ENOMEM;
        }
	if (usb_set_interface (s->usbdev, _DABUSB_IF, 1) < 0) {
		dev_err(&s->usbdev->dev, "set_interface failed");
		return -EINVAL;
	}
        if (dabusb_alloc_buffers (s) < 0)
		return -ENOMEM;
	if (usb_set_interface (s->usbdev, _DABUSB_IF, 0) < 0) {
		dev_err(&s->usbdev->dev, "set_interface failed");
		return -EINVAL;
	}
        
	dev_err(&s->usbdev->dev, "Now attached to dabusb%i, %p\n",s->devnum-DABUSB_MINOR,s);
        return 0;

probe_failed:
	kfree(s);
	return retval;
}

static void dabusb_disconnect (struct usb_interface *intf)
{
	wait_queue_t __wait;
	pdabusb_t s = usb_get_intfdata (intf);

	dbg("dabusb_disconnect");
	down (&disconnect_sem);

	init_waitqueue_entry(&__wait, current);
	usb_set_intfdata (intf, NULL);
	
	if (s) {
		usb_deregister_dev (intf, &dabusb_class);
		s->remove_pending = 1;
		wake_up (&s->wait);
		add_wait_queue(&s->remove_ok, &__wait);
		set_current_state(TASK_INTERRUPTIBLE);
		if (s->state == _started || s->opened)
			schedule();
		current->state = TASK_RUNNING;
		remove_wait_queue(&s->remove_ok, &__wait);
		dabusb_free_buffers (s);
		s->usbdev = NULL;
		s->overruns = 0;
		kfree(s);
	}
	up (&disconnect_sem);
}


/* --------------------------------------------------------------------- */

static int __init dabusb_init (void)
{

	dbg("dabusb_init: driver registered");
	printk(KERN_INFO DRIVER_VERSION ":" DRIVER_DESC);

	/* register misc device */
	if (usb_register(&dabusb_driver))
		return -1;

	return 0;
}

static void __exit dabusb_cleanup (void)
{
	dbg("dabusb_cleanup");

	usb_deregister (&dabusb_driver);
}

/* --------------------------------------------------------------------- */

MODULE_AUTHOR( DRIVER_AUTHOR );
MODULE_DESCRIPTION( DRIVER_DESC );
MODULE_LICENSE("GPL");

module_param(buffers, int, 0);
MODULE_PARM_DESC (buffers, "Number of buffers (default=256)");

module_init (dabusb_init);
module_exit (dabusb_cleanup);

/* --------------------------------------------------------------------- */
